import copy

import abjad


def test_NoteHead___cmp___01():
    note_head_1 = abjad.NoteHead(12)
    note_head_2 = abjad.NoteHead(12)

    assert not note_head_1 < note_head_2
    assert note_head_1 <= note_head_2
    assert note_head_1 == note_head_2
    assert not note_head_1 != note_head_2
    assert not note_head_1 > note_head_2
    assert note_head_1 >= note_head_2


def test_NoteHead___cmp___02():
    note_head_1 = abjad.NoteHead(12)
    note_head_2 = abjad.NoteHead(13)

    assert not note_head_2 < note_head_1
    assert not note_head_2 <= note_head_1
    assert not note_head_2 == note_head_1
    assert note_head_2 != note_head_1
    assert note_head_2 > note_head_1
    assert note_head_2 >= note_head_1


def test_NoteHead___cmp___03():
    note_head_1 = abjad.NoteHead(12)
    note_head_2 = abjad.NoteHead("c''")

    assert not note_head_1 < note_head_2
    assert note_head_1 <= note_head_2
    assert note_head_1 == note_head_2
    assert not note_head_1 != note_head_2
    assert not note_head_1 > note_head_2
    assert note_head_1 >= note_head_2


def test_NoteHead___cmp___04():
    note_head_1 = abjad.NoteHead(12)
    note_head_2 = 13

    assert not note_head_2 < note_head_1
    assert not note_head_2 <= note_head_1
    assert not note_head_2 == note_head_1
    assert note_head_2 != note_head_1
    assert note_head_2 > note_head_1
    assert note_head_2 >= note_head_1


def test_NoteHead___copy___01():
    note_head_1 = abjad.NoteHead("cs''")
    note_head_1.is_cautionary = True
    note_head_1.is_forced = True
    abjad.tweak(note_head_1, r"\tweak color #red")
    abjad.tweak(note_head_1, r"\tweak font-size -2")

    note_head_2 = copy.copy(note_head_1)

    assert isinstance(note_head_1, abjad.NoteHead)
    assert isinstance(note_head_2, abjad.NoteHead)
    assert note_head_1 == note_head_2
    assert note_head_1 is not note_head_2
    assert note_head_1.is_cautionary == note_head_2.is_cautionary
    assert note_head_1.is_forced == note_head_2.is_forced
    assert note_head_1.tweaks == note_head_2.tweaks
    assert note_head_1.tweaks is not note_head_2.tweaks


def test_NoteHead___deepcopy___01():
    note_head_1 = abjad.NoteHead("cs''")
    abjad.tweak(note_head_1, r"\tweak color #red")
    note_head_1.is_cautionary = True
    note_head_1.is_forced = True

    note_head_2 = copy.deepcopy(note_head_1)

    assert isinstance(note_head_1, abjad.NoteHead)
    assert isinstance(note_head_2, abjad.NoteHead)
    assert note_head_1 == note_head_2
    assert note_head_1 is not note_head_2
    assert note_head_1.is_cautionary == note_head_2.is_cautionary
    assert note_head_1.is_forced == note_head_2.is_forced
    assert note_head_1.tweaks == note_head_2.tweaks
    assert note_head_1.tweaks is not note_head_2.tweaks


def test_NoteHead___init___01():
    """
    Initialize note-head by number.
    """

    notehead = abjad.NoteHead(6)
    assert notehead.written_pitch == abjad.NamedPitch(6)


def test_NoteHead___init___02():
    """
    Initialize note-head by LilyPond-style pitch string.
    """

    notehead = abjad.NoteHead("cs,,,")
    assert notehead.written_pitch == abjad.NamedPitch("cs,,,")


def test_NoteHead___init___03():
    """
    Initialize note-head by other note-head instance.
    """

    notehead = abjad.NoteHead(6)
    new = abjad.NoteHead(notehead)

    assert notehead is not new
    assert notehead.written_pitch == "fs'"
    assert new.written_pitch != 6


def test_NoteHead___init___04():
    """
    Initialize note-head with tweak manager.
    """

    note_head = abjad.NoteHead("cs''", tweaks=[abjad.Tweak(r"\tweak color #red")])

    assert abjad.lilypond(note_head) == "\\tweak color #red\ncs''"


def test_NoteHead_is_forced_01():
    note_head = abjad.NoteHead(written_pitch="c'")
    assert note_head.is_forced is None
    note_head.is_forced = True
    assert note_head.is_forced is True
    note_head.is_forced = False
    assert note_head.is_forced is False


def test_NoteHead_is_parenthesized_01():
    note_head = abjad.NoteHead(written_pitch="c'")
    assert note_head.is_parenthesized is None
    note_head.is_parenthesized = True
    assert note_head.is_parenthesized is True
    note_head.is_parenthesized = False
    assert note_head.is_parenthesized is False


def test_NoteHead_is_parenthesized_02():
    note_head = abjad.NoteHead(written_pitch="c'")
    note_head.is_parenthesized = True
    assert abjad.lilypond(note_head) == abjad.string.normalize(
        r"""
        \parenthesize
        c'
        """
    )


def test_NoteHead_is_parenthesized_03():
    note = abjad.Note("c'4")
    note.note_head.is_parenthesized = True
    assert abjad.lilypond(note) == abjad.string.normalize(
        r"""
        \parenthesize
        c'4
        """
    )


def test_NoteHead_is_parenthesized_04():
    chord = abjad.Chord("<c' e' g'>4")
    chord.note_heads[1].is_parenthesized = True
    assert abjad.lilypond(chord) == abjad.string.normalize(
        r"""
        <
            c'
            \parenthesize
            e'
            g'
        >4
        """
    )


def test_NoteHead_written_pitch_01():
    """
    Set note-head head pitch with integer.
    """

    note = abjad.Note(13, (1, 4))
    note.note_head.written_pitch = 14

    "NoteHead(d'')"

    assert abjad.lilypond(note.note_head) == "d''"
    assert note.note_head.written_pitch != 14


def test_NoteHead_written_pitch_02():
    """
    Set note-head pitch with pitch.
    """

    note = abjad.Note(13, (1, 4))
    note.note_head.written_pitch = abjad.NamedPitch(14)

    "NoteHead(d'')"

    assert abjad.lilypond(note.note_head) == "d''"
    assert note.note_head.written_pitch != 14


def test_NoteHead_written_pitch_03():
    """
    Set note-head pitch from another note or note-head.
    Make sure this does not cause reference problems.
    """

    n1 = abjad.Note(12, (1, 4))
    n2 = abjad.Note(14, (1, 4))
    n1.written_pitch = n2.written_pitch

    assert n1.written_pitch == abjad.NamedPitch(14)
    assert n2.written_pitch == abjad.NamedPitch(14)
    assert n1.written_pitch is not n2.written_pitch



import collections
import fractions
import math
import numbers

from . import _getlib
from . import duration as _duration
from . import exceptions as _exceptions
from . import math as _math
from . import pitch as _pitch
from . import score as _score
from . import sequence as _sequence
from . import spanners as _spanners
from . import tag as _tag
from . import typings as _typings


def _group_by_implied_prolation(durations):
    pairs = []
    for duration in durations:
        if isinstance(duration, tuple):
            pairs.append(duration)
        else:
            pairs.append(duration.pair)
    durations = pairs
    assert 0 < len(durations)
    group = [durations[0]]
    result = [group]
    for d in durations[1:]:
        d_f = set(_math.factors(d[1]))
        d_f.discard(2)
        gd_f = set(_math.factors(group[0][1]))
        gd_f.discard(2)
        if d_f == gd_f:
            group.append(d)
        else:
            group = [d]
            result.append(group)
    return result


def _make_leaf_on_pitch(
    pitch,
    duration,
    *,
    increase_monotonic=None,
    forbidden_note_duration=None,
    forbidden_rest_duration=None,
    skips_instead_of_rests=None,
    tag=None,
    use_multimeasure_rests=None,
):
    note_prototype = (
        numbers.Number,
        str,
        _pitch.NamedPitch,
        _pitch.NumberedPitch,
        _pitch.PitchClass,
    )
    chord_prototype = (tuple, list)
    rest_prototype = (type(None),)
    if isinstance(pitch, note_prototype):
        leaves = _make_tied_leaf(
            _score.Note,
            duration,
            increase_monotonic=increase_monotonic,
            forbidden_duration=forbidden_note_duration,
            pitches=pitch,
            tag=tag,
        )
    elif isinstance(pitch, chord_prototype):
        leaves = _make_tied_leaf(
            _score.Chord,
            duration,
            increase_monotonic=increase_monotonic,
            forbidden_duration=forbidden_note_duration,
            pitches=pitch,
            tag=tag,
        )
    elif isinstance(pitch, rest_prototype) and skips_instead_of_rests:
        leaves = _make_tied_leaf(
            _score.Skip,
            duration,
            increase_monotonic=increase_monotonic,
            forbidden_duration=forbidden_rest_duration,
            pitches=None,
            tag=tag,
        )
    elif isinstance(pitch, rest_prototype) and not use_multimeasure_rests:
        leaves = _make_tied_leaf(
            _score.Rest,
            duration,
            increase_monotonic=increase_monotonic,
            forbidden_duration=forbidden_rest_duration,
            pitches=None,
            tag=tag,
        )
    elif isinstance(pitch, rest_prototype) and use_multimeasure_rests:
        multimeasure_rest = _score.MultimeasureRest((1), tag=tag)
        multimeasure_rest.multiplier = duration
        leaves = (multimeasure_rest,)
    else:
        raise ValueError(f"unknown pitch: {pitch!r}.")
    return leaves


def _make_tied_leaf(
    class_,
    duration,
    increase_monotonic=None,
    forbidden_duration=None,
    multiplier=None,
    pitches=None,
    tag=None,
    tie_parts=True,
):
    duration = _duration.Duration(duration)
    duration_pair = duration.pair
    if forbidden_duration is not None:
        assert forbidden_duration.is_assignable
        assert forbidden_duration.numerator == 1
    # find preferred numerator of written durations if necessary
    if forbidden_duration is not None and forbidden_duration <= fractions.Fraction(
        *duration_pair
    ):
        denominators = [
            2 * forbidden_duration.denominator,
            duration_pair[1],
        ]
        denominator = _math.least_common_multiple(*denominators)
        pair = _duration.with_denominator(forbidden_duration, denominator)
        forbidden_numerator = pair[0]
        assert forbidden_numerator % 2 == 0
        preferred_numerator = forbidden_numerator / 2
        pair = _duration.with_denominator(duration_pair, denominator)
        duration_pair = pair
    # make written duration numerators
    numerators = []
    parts = _math.partition_integer_into_canonic_parts(duration_pair[0])
    if forbidden_duration is not None and fractions.Fraction(
        forbidden_duration
    ) <= fractions.Fraction(*duration_pair):
        for part in parts:
            if forbidden_numerator <= part:
                better_parts = _partition_less_than_double(part, preferred_numerator)
                numerators.extend(better_parts)
            else:
                numerators.append(part)
    else:
        numerators = parts
    # reverse numerators if necessary
    if increase_monotonic:
        numerators = list(reversed(numerators))
    # make one leaf per written duration
    result = []
    for numerator in numerators:
        written_duration = _duration.Duration(numerator, duration_pair[1])
        if pitches is not None:
            arguments = (pitches, written_duration)
        else:
            arguments = (written_duration,)
        result.append(class_(*arguments, multiplier=multiplier, tag=tag))
    # tie if required
    if tie_parts and 1 < len(result):
        if not issubclass(class_, (_score.Rest, _score.Skip)):
            _spanners.tie(result)
    return result


def _make_unprolated_notes(pitches, durations, increase_monotonic=None, tag=None):
    assert len(pitches) == len(durations)
    result = []
    for pitch, duration in zip(pitches, durations):
        result.extend(
            _make_tied_leaf(
                _score.Note,
                duration,
                pitches=pitch,
                increase_monotonic=increase_monotonic,
                tag=tag,
            )
        )
    return result


def _partition_less_than_double(n, m):
    assert _math.is_positive_integer_equivalent_number(n)
    assert _math.is_positive_integer_equivalent_number(m)
    n, m = int(n), int(m)
    result = []
    current_value = n
    double_m = 2 * m
    while double_m <= current_value:
        result.append(m)
        current_value -= m
    result.append(current_value)
    return tuple(result)


def make_leaves(
    pitches,
    durations,
    *,
    forbidden_note_duration: _typings.Duration | None = None,
    forbidden_rest_duration: _typings.Duration | None = None,
    skips_instead_of_rests: bool = False,
    increase_monotonic: bool = False,
    tag: _tag.Tag | None = None,
    use_multimeasure_rests: bool = False,
):
    r"""
    Makes leaves from ``pitches`` and ``durations``.

    ..  container:: example

        Integer and string elements in ``pitches`` result in notes:

        >>> pitches = [2, 4, "F#5", "G#5"]
        >>> duration = abjad.Duration(1, 4)
        >>> leaves = abjad.makers.make_leaves(pitches, duration)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                d'4
                e'4
                fs''4
                gs''4
            }

    ..  container:: example

        Tuple elements in ``pitches`` result in chords:

        >>> pitches = [(0, 2, 4), ("F#5", "G#5", "A#5")]
        >>> duration = abjad.Duration(1, 2)
        >>> leaves = abjad.makers.make_leaves(pitches, duration)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <c' d' e'>2
                <fs'' gs'' as''>2
            }

    ..  container:: example

        None-valued elements in ``pitches`` result in rests:

        >>> pitches = 4 * [None]
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves, lilypond_type="RhythmicStaff")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new RhythmicStaff
            {
                r4
                r4
                r4
                r4
            }

    ..  container:: example

        You can mix and match values passed to ``pitches``:

        >>> pitches = [(0, 2, 4), None, "C#5", "D#5"]
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <c' d' e'>4
                r4
                cs''4
                ds''4
            }

    ..  container:: example

        Works with segments:

        >>> pitches = "e'' ef'' d'' df'' c''"
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                e''4
                ef''4
                d''4
                df''4
                c''4
            }

    ..  container:: example

        Reads ``pitches`` cyclically when the length of ``pitches`` is less than the
        length of ``durations``:

        >>> pitches = ["C5"]
        >>> durations = 2 * [abjad.Duration(3, 8), abjad.Duration(1, 8)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c''4.
                c''8
                c''4.
                c''8
            }

    ..  container:: example

        Reads ``durations`` cyclically when the length of ``durations`` is less than the
        length of ``pitches``:

        >>> pitches = "c'' d'' e'' f''"
        >>> durations = [abjad.Duration(1, 4)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c''4
                d''4
                e''4
                f''4
            }

    ..  container:: example

        Elements in ``durations`` with non-power-of-two denominators result in
        tuplet-nested leaves:

        >>> pitches = ["D5"]
        >>> durations = 3 * [abjad.Duration(1, 3)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \tuplet 3/2
                {
                    d''2
                    d''2
                    d''2
                }
            }

    ..  container:: example

        Set ``increase_monotonic`` to false to return nonassignable durations tied from
        greatest to least:

        >>> pitches = ["D#5"]
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> time_signature = abjad.TimeSignature((13, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 13/16
                ds''2.
                ~
                ds''16
            }

    ..  container:: example

        Set ``increase_monotonic`` to true to return nonassignable durations tied from
        least to greatest:

        >>> pitches = ["E5"]
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations, increase_monotonic=True)
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> time_signature = abjad.TimeSignature((13, 16))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 13/16
                e''16
                ~
                e''2.
            }

    ..  container:: example

        Set ``forbidden_note_duration`` to avoid notes greater than or equal to a certain
        written duration:

        >>> pitches = "f' g'"
        >>> durations = [abjad.Duration(5, 8)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations,
        ...     forbidden_note_duration=abjad.Duration(1, 2))
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> time_signature = abjad.TimeSignature((5, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 5/4
                f'4
                ~
                f'4
                ~
                f'8
                g'4
                ~
                g'4
                ~
                g'8
            }

    ..  container:: example

        You may set ``forbidden_note_duration`` and ``increase_monotonic`` together:

        >>> pitches = "f' g'"
        >>> durations = [abjad.Duration(5, 8)]
        >>> leaves = abjad.makers.make_leaves(
        ...     pitches, durations,
        ...     forbidden_note_duration=abjad.Duration(1, 2),
        ...     increase_monotonic=True,
        ... )
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> time_signature = abjad.TimeSignature((5, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 5/4
                f'8
                ~
                f'4
                ~
                f'4
                g'8
                ~
                g'4
                ~
                g'4
            }

    ..  container:: example

        Produces diminished tuplets:

        >>> pitches = "f'"
        >>> durations = [abjad.Duration(5, 14)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> time_signature = abjad.TimeSignature((5, 14))
        >>> leaf = abjad.get.leaf(staff, 0)
        >>> abjad.attach(time_signature, leaf)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \tweak edge-height #'(0.7 . 0)
                \tuplet 14/8
                {
                    #(ly:expect-warning "strange time signature found")
                    \time 5/14
                    f'2
                    ~
                    f'8
                }
            }

        This is default behavior.

    ..  container:: example

        None-valued elements in ``pitches`` result in multimeasure rests when the
        multimeasure rest keyword is set:

        >>> pitches = [None]
        >>> durations = [abjad.Duration(3, 8), abjad.Duration(5, 8)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations,
        ...     use_multimeasure_rests=True)
        >>> leaves
        [MultimeasureRest('R1 * 3/8'), MultimeasureRest('R1 * 5/8')]

        >>> staff = abjad.Staff(leaves, lilypond_type="RhythmicStaff")
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 8)), leaves[0])
        >>> abjad.attach(abjad.TimeSignature((5, 8)), leaves[1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new RhythmicStaff
            {
                \time 3/8
                R1 * 3/8
                \time 5/8
                R1 * 5/8
            }

    ..  container:: example

        Works with numbered pitch-class:

        >>> pitches = [abjad.NumberedPitchClass(6)]
        >>> durations = [abjad.Duration(13, 16)]
        >>> leaves = abjad.makers.make_leaves(pitches, durations)
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                fs'2.
                ~
                fs'16
            }

    ..  container:: example

        Makes skips instead of rests:

        >>> pitches = [None]
        >>> durations = [abjad.Duration(13, 16)]
        >>> abjad.makers.make_leaves(pitches, durations, skips_instead_of_rests=True)
        [Skip('s2.'), Skip('s16')]

    ..  container:: example

        Integer and string elements in ``pitches`` result in notes:

        >>> tag = abjad.Tag("leaf_maker")
        >>> pitches = [2, 4, "F#5", "G#5"]
        >>> duration = abjad.Duration(1, 4)
        >>> leaves = abjad.makers.make_leaves(pitches, duration, tag=tag)
        >>> staff = abjad.Staff(leaves)
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! leaf_maker
                d'4
                %! leaf_maker
                e'4
                %! leaf_maker
                fs''4
                %! leaf_maker
                gs''4
            }

    """
    if isinstance(pitches, str):
        pitches = pitches.split()
    if not isinstance(pitches, collections.abc.Iterable):
        pitches = [pitches]
    if isinstance(durations, numbers.Number | tuple):
        durations = [durations]
    if forbidden_note_duration is not None:
        forbidden_note_duration = _duration.Duration(forbidden_note_duration)
    if forbidden_rest_duration is not None:
        forbidden_rest_duration = _duration.Duration(forbidden_rest_duration)
    nonreduced_fractions = [_duration.Duration(_) for _ in durations]
    size = max(len(nonreduced_fractions), len(pitches))
    nonreduced_fractions = _sequence.repeat_to_length(nonreduced_fractions, size)
    pitches = _sequence.repeat_to_length(pitches, size)
    duration_groups = _group_by_implied_prolation(nonreduced_fractions)
    result: list[_score.Tuplet | _score.Leaf] = []
    for duration_group in duration_groups:
        factors_ = _math.factors(duration_group[0][1])
        factors = set(factors_)
        factors.discard(1)
        factors.discard(2)
        current_pitches = pitches[0 : len(duration_group)]
        pitches = pitches[len(duration_group) :]
        if len(factors) == 0:
            for pitch, duration in zip(current_pitches, duration_group):
                leaves = _make_leaf_on_pitch(
                    pitch,
                    duration,
                    increase_monotonic=increase_monotonic,
                    forbidden_note_duration=forbidden_note_duration,
                    forbidden_rest_duration=forbidden_rest_duration,
                    skips_instead_of_rests=skips_instead_of_rests,
                    tag=tag,
                    use_multimeasure_rests=use_multimeasure_rests,
                )
                result.extend(leaves)
        else:
            denominator = duration_group[0][1]
            numerator = _math.greatest_power_of_two_less_equal(denominator)
            multiplier = (numerator, denominator)
            ratio = 1 / _duration.Duration(*multiplier)
            duration_group = [
                ratio * _duration.Duration(duration) for duration in duration_group
            ]
            tuplet_leaves: list[_score.Leaf] = []
            for pitch, duration in zip(current_pitches, duration_group):
                leaves = _make_leaf_on_pitch(
                    pitch,
                    duration,
                    increase_monotonic=increase_monotonic,
                    skips_instead_of_rests=skips_instead_of_rests,
                    tag=tag,
                    use_multimeasure_rests=use_multimeasure_rests,
                )
                tuplet_leaves.extend(leaves)
            tuplet = _score.Tuplet(multiplier, tuplet_leaves)
            result.append(tuplet)
    return result


def make_notes(
    pitches, durations, *, increase_monotonic: bool = False, tag: _tag.Tag | None = None
) -> list[_score.Note | _score.Tuplet]:
    r"""
    Makes notes from ``pitches`` and ``durations``.

    Set ``pitches`` to a single pitch or a sequence of pitches.

    Set ``durations`` to a single duration or a list of durations.

    ..  container:: example

        Cycles through ``pitches`` when the length of ``pitches`` is less than
        the length of ``durations``:

        >>> pitches = [0]
        >>> durations = [(1, 16), (1, 8), (1, 8)]
        >>> notes = abjad.makers.make_notes(pitches, durations)
        >>> staff = abjad.Staff(notes)
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'16
                c'8
                c'8
            }

    ..  container:: example

        Cycles through ``durations`` when the length of ``durations`` is less than the
        length of ``pitches``:

        >>> pitches = [0, 2, 4, 5, 7]
        >>> durations = [(1, 16), (1, 8), (1, 8)]
        >>> notes = abjad.makers.make_notes(pitches, durations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'16
                d'8
                e'8
                f'16
                g'8
            }

    ..  container:: example

        Creates ad hoc tuplets for nonassignable durations:

        >>> pitches = [0]
        >>> durations = [(1, 16), (1, 12), (1, 8)]
        >>> notes = abjad.makers.make_notes(pitches, durations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'16
                \tweak edge-height #'(0.7 . 0)
                \tuplet 3/2
                {
                    c'8
                }
                c'8
            }

    ..  container:: example

        Tied values are written in decreasing duration by default:

        >>> pitches = [0]
        >>> durations = [(13, 16)]
        >>> notes = abjad.makers.make_notes(pitches, durations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2.
                ~
                c'16
            }

        Set ``increase_monotonic=True`` to express tied values in increasing duration:

        >>> pitches = [0]
        >>> durations = [(13, 16)]
        >>> notes = abjad.makers.make_notes(pitches, durations, increase_monotonic=True)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'16
                ~
                c'2.
            }

    ..  container:: example

        Works with pitch segments:

        >>> pitches = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])
        >>> durations = [(1, 8)]
        >>> notes = abjad.makers.make_notes(pitches, durations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                bf8
                bqf8
                fs'8
                g'8
                bqf8
                g'8
            }

    ..  container:: example

        Tag output like this:

        >>> pitches = [0]
        >>> durations = [(1, 16), (1, 8), (1, 8)]
        >>> tag = abjad.Tag("note_maker")
        >>> notes = abjad.makers.make_notes(pitches, durations, tag=tag)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! note_maker
                c'16
                %! note_maker
                c'8
                %! note_maker
                c'8
            }

    """
    if isinstance(pitches, str):
        pitches = pitches.split()
    if not isinstance(pitches, collections.abc.Iterable):
        pitches = [pitches]
    if isinstance(durations, numbers.Number | tuple):
        durations = [durations]
    pairs = []
    for duration in durations:
        if isinstance(duration, tuple):
            pairs.append(duration)
        elif isinstance(duration, int):
            pair = (duration, 1)
            pairs.append(pair)
        else:
            pairs.append(duration.pair)
    size = max(len(pairs), len(pitches))
    pairs = _sequence.repeat_to_length(pairs, size)
    pitches = _sequence.repeat_to_length(pitches, size)
    durations = _group_by_implied_prolation(pairs)
    result: list[_score.Note | _score.Tuplet] = []
    for duration in durations:
        factors = set(_math.factors(duration[0][1]))
        factors.discard(1)
        factors.discard(2)
        ps = pitches[0 : len(duration)]
        pitches = pitches[len(duration) :]
        if len(factors) == 0:
            result.extend(
                _make_unprolated_notes(
                    ps,
                    duration,
                    increase_monotonic=increase_monotonic,
                    tag=tag,
                )
            )
        else:
            denominator = duration[0][1]
            numerator = _math.greatest_power_of_two_less_equal(denominator)
            multiplier = _duration.Duration(numerator, denominator)
            ratio = multiplier.reciprocal
            duration = [ratio * _duration.Duration(d) for d in duration]
            ns = _make_unprolated_notes(
                ps,
                duration,
                increase_monotonic=increase_monotonic,
                tag=tag,
            )
            tuplet = _score.Tuplet(multiplier.pair, ns)
            result.append(tuplet)
    return result


def tuplet_from_duration_and_ratio(
    duration,
    ratio: tuple[int, ...],
    *,
    increase_monotonic: bool = False,
    tag: _tag.Tag | None = None,
) -> _score.Tuplet:
    r"""
    Makes tuplet from ``duration`` and ``ratio``.

    ..  container:: example

        Makes tupletted leaves strictly without dots when all ``ratio`` equal ``1``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, 1, 1, -1, -1),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'32.
                c'32.
                r32.
                r32.
            }

        Allows tupletted leaves to return with dots when some ``ratio`` do not equal
        ``1``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, -2, -2, 3, 3),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 11/6
            {
                \time 3/16
                c'32
                r16
                r16
                c'16.
                c'16.
            }

        Interprets nonassignable ``ratio`` according to ``increase_monotonic``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (5, -1, 5),
        ...     increase_monotonic=True,
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 11/8
            {
                \time 3/16
                c'16...
                r64.
                c'16...
            }

    ..  container:: example

        Makes augmented tuplet from ``duration`` and ``ratio`` and encourages dots:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, 1, 1, -1, -1),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'32.
                c'32.
                r32.
                r32.
            }

        Interprets nonassignable ``ratio`` according to ``increase_monotonic``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (5, -1, 5),
        ...     increase_monotonic=True,
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 11/8
            {
                \time 3/16
                c'16...
                r64.
                c'16...
            }

    ..  container:: example

        Makes diminished tuplet from ``duration`` and nonzero integer ``ratio``.

        Makes tupletted leaves strictly without dots when all ``ratio`` equal ``1``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, 1, 1, -1, -1),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'32.
                c'32.
                r32.
                r32.
            }

        Allows tupletted leaves to return with dots when some ``ratio`` do not equal
        ``1``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, -2, -2, 3, 3),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 11/6
            {
                \time 3/16
                c'32
                r16
                r16
                c'16.
                c'16.
            }

        Interprets nonassignable ``ratio`` according to ``increase_monotonic``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (5, -1, 5),
        ...     increase_monotonic=True,
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 11/8
            {
                \time 3/16
                c'16...
                r64.
                c'16...
            }

    ..  container:: example

        Makes diminished tuplet from ``duration`` and ``ratio`` and encourages dots:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (1, 1, 1, -1, -1),
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'32.
                c'32.
                r32.
                r32.
            }

        Interprets nonassignable ``ratio`` according to ``direction``:

        >>> tuplet = abjad.makers.tuplet_from_duration_and_ratio(
        ...     abjad.Duration(3, 16),
        ...     (5, -1, 5),
        ...     increase_monotonic=True,
        ... )
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 11/8
            {
                \time 3/16
                c'16...
                r64.
                c'16...
            }

    Reduces ``ratio`` relative to each other.

    Interprets negative ``ratio`` as rests.
    """
    duration = _duration.Duration(duration)
    assert isinstance(ratio, tuple), repr(ratio)
    basic_prolated_duration = duration / _math.weight(ratio)
    basic_written_duration = basic_prolated_duration.equal_or_greater_assignable
    written_durations = [x * basic_written_duration for x in ratio]
    notes: list[_score.Leaf | _score.Tuplet]
    try:
        notes = [
            _score.Note(0, x, tag=tag) if 0 < x else _score.Rest(abs(x), tag=tag)
            for x in written_durations
        ]
    except _exceptions.AssignabilityError:
        denominator = duration.denominator
        note_durations = [_duration.Duration(x, denominator) for x in ratio]
        pitches = [None if note_duration < 0 else 0 for note_duration in note_durations]
        leaf_durations = [abs(note_duration) for note_duration in note_durations]
        notes = make_leaves(
            pitches,
            leaf_durations,
            increase_monotonic=increase_monotonic,
            tag=tag,
        )
    tuplet = _score.Tuplet.from_duration(duration, notes, tag=tag)
    tuplet.normalize_multiplier()
    return tuplet


def tuplet_from_leaf_and_ratio(
    leaf: _score.Leaf, ratio: tuple[int, ...]
) -> _score.Tuplet:
    r"""
    Makes tuplet from ``leaf`` and ``ratio``.

    >>> note = abjad.Note("c'8.")

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1,))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 3/16
                c'8.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 3/16
                c'16
                c'8
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'16.
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 4/3
            {
                \time 3/16
                c'32
                c'16
                c'16
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3, 3))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 11/6
            {
                \time 3/16
                c'32
                c'16
                c'16
                c'16.
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3, 3, 4))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'64
                c'32
                c'32
                c'32.
                c'32.
                c'16
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1,))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 3/16
                c'8.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 3/16
                c'16
                c'8
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'32.
                c'16.
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 4/3
            {
                \time 3/16
                c'32
                c'16
                c'16
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3, 3))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 11/6
            {
                \time 3/16
                c'32
                c'16
                c'16
                c'16.
                c'16.
            }

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_leaf_and_ratio(note, (1, 2, 2, 3, 3, 4))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(tuplet)
            >>> print(string)
            \tuplet 5/4
            {
                \time 3/16
                c'64
                c'32
                c'32
                c'32.
                c'32.
                c'16
            }

    """
    assert isinstance(ratio, tuple), repr(ratio)
    target_duration = leaf.written_duration
    basic_prolated_duration = target_duration / sum(ratio)
    basic_written_duration = basic_prolated_duration.equal_or_greater_assignable
    written_durations = [_ * basic_written_duration for _ in ratio]
    notes: list[_score.Note | _score.Tuplet]
    try:
        notes = [_score.Note(0, x) for x in written_durations]
    except _exceptions.AssignabilityError:
        denominator = target_duration.denominator
        note_durations = [_duration.Duration(_, denominator) for _ in ratio]
        notes = make_notes(0, note_durations)
    contents_duration = _getlib._get_duration(notes)
    multiplier = target_duration / contents_duration
    tuplet = _score.Tuplet(_duration.pair(multiplier), notes)
    tuplet.normalize_multiplier()
    return tuplet


def tuplet_from_ratio_and_pair(
    ratio: tuple,
    fraction: tuple,
    *,
    tag: _tag.Tag | None = None,
) -> _score.Tuplet:
    r"""
    Makes tuplet from nonreduced ``ratio`` and nonreduced ``fraction``.

    ..  container:: example

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1,), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 7/16
                c'4..
            }

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 6/7
            {
                \time 7/16
                c'8
                c'4
            }

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 4), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 1/1
            {
                \time 7/16
                c'16
                c'8
                c'4
            }

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 4, 1), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 8/7
            {
                \time 7/16
                c'16
                c'8
                c'4
                c'16
            }

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 4, 1, 2), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 10/7
            {
                \time 7/16
                c'16
                c'8
                c'4
                c'16
                c'8
            }

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((1, 2, 4, 1, 2, 4), (7, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tuplet 2/1
            {
                \time 7/16
                c'16
                c'8
                c'4
                c'16
                c'8
                c'4
            }

        Works with nonassignable rests:

        >>> tuplet = abjad.makers.tuplet_from_ratio_and_pair((11, -5), (5, 16))
        >>> staff = abjad.Staff(
        ...     [tuplet],
        ...     lilypond_type="RhythmicStaff",
        ... )
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((7, 16)), tuplet[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff[0])
            >>> print(string)
            \tweak text #tuplet-number::calc-fraction-text
            \tuplet 8/5
            {
                \time 7/16
                c'4
                ~
                c'16.
                r8
                r32
            }

    Interprets ``d`` as tuplet denominator.
    """
    if not isinstance(ratio, tuple):
        raise ValueError(f"must be tuple, not {ratio!r}.")
    if not isinstance(fraction, tuple):
        raise ValueError(f"must be pair, not {fraction!r}.")
    numerator, denominator = fraction
    duration = _duration.Duration(fraction)
    if len(ratio) == 1:
        if 0 < ratio[0]:
            try:
                note = _score.Note(0, duration, tag=tag)
                duration = note._get_duration()
                tuplet = _score.Tuplet.from_duration(duration, [note], tag=tag)
                return tuplet
            except _exceptions.AssignabilityError:
                notes = make_notes(0, duration, tag=tag)
                duration = _getlib._get_duration(notes)
                return _score.Tuplet.from_duration(duration, notes, tag=tag)
        elif ratio[0] < 0:
            try:
                rest = _score.Rest(duration, tag=tag)
                duration = rest._get_duration()
                return _score.Tuplet.from_duration(duration, [rest], tag=tag)
            except _exceptions.AssignabilityError:
                rests = make_leaves([None], duration, tag=tag)
                duration = _getlib._get_duration(rests)
                return _score.Tuplet.from_duration(duration, rests, tag=tag)
        else:
            raise ValueError("no divide zero values.")
    else:
        exponent = int(math.log(_math.weight(ratio), 2) - math.log(numerator, 2))
        denominator = int(denominator * 2**exponent)
        components: list[_score.Leaf | _score.Tuplet] = []
        for x in ratio:
            if not x:
                raise ValueError("no divide zero values.")
            if 0 < x:
                try:
                    note = _score.Note(0, (x, denominator), tag=tag)
                    components.append(note)
                except _exceptions.AssignabilityError:
                    notes = make_notes(0, (x, denominator), tag=tag)
                    components.extend(notes)
            else:
                try:
                    rest = _score.Rest((-x, denominator), tag=tag)
                    components.append(rest)
                except _exceptions.AssignabilityError:
                    rests = make_leaves(None, (-x, denominator), tag=tag)
                    components.extend(rests)
        return _score.Tuplet.from_duration(duration, components, tag=tag)



import functools
import itertools
import typing

from . import math as _math
from . import sequence as _sequence


def _is_restricted_growth_function(sequence):
    """
    Is true when ``sequence`` is a restricted growth function.

    ..  container:: example

        Is true when sequence is a restricted growth function:

        >>> abjad.enumerate._is_restricted_growth_function([1, 1, 1, 1])
        True

        >>> abjad.enumerate._is_restricted_growth_function([1, 1, 1, 2])
        True

        >>> abjad.enumerate._is_restricted_growth_function([1, 1, 2, 1])
        True

        >>> abjad.enumerate._is_restricted_growth_function([1, 1, 2, 2])
        True

    ..  container:: example

        Is false when sequence is not a restricted growth function:

        >>> abjad.enumerate._is_restricted_growth_function([1, 1, 1, 3])
        False

        >>> abjad.enumerate._is_restricted_growth_function([17])
        False

    A restricted growth function is a sequence ``l`` such that ``l[0] == 1`` and such
    that ``l[i] <= max(l[:i]) + 1`` for ``1 <= i <= len(l)``.

    Returns true or false.
    """
    try:
        for i, n in enumerate(sequence):
            if i == 0:
                if not n == 1:
                    return False
            else:
                if not n <= max(sequence[:i]) + 1:
                    return False
        return True
    except TypeError:
        return False


def _partition_by_rgf(sequence, rgf):
    """
    Partitions ``sequence`` by restricted growth function ``rgf``.

    >>> sequence = list(range(10))
    >>> rgf = [1, 1, 2, 2, 1, 2, 3, 3, 2, 4]

    >>> abjad.enumerate._partition_by_rgf(sequence, rgf)
    [[0, 1, 4], [2, 3, 5, 8], [6, 7], [9]]

    Returns list of lists.
    """
    rgf = list(rgf)
    if not _is_restricted_growth_function(rgf):
        raise ValueError(f"must be restricted growth function: {rgf!r}.")
    if not len(sequence) == len(rgf):
        raise ValueError("lengths must be equal.")
    partition = []
    for part_index in range(max(rgf)):
        part = []
        partition.append(part)
    for n, part_number in zip(sequence, rgf):
        part_index = part_number - 1
        part = partition[part_index]
        part.append(n)
    partition = [type(sequence)(_) for _ in partition]
    return type(sequence)(partition)


def _yield_restricted_growth_functions(length):
    """
    Yields restricted growth functions of ``length``.

    ..  container:: example

        >>> rgfs = abjad.enumerate._yield_restricted_growth_functions(4)
        >>> for rgf in rgfs:
        ...     rgf
        ...
        (1, 1, 1, 1)
        (1, 1, 1, 2)
        (1, 1, 2, 1)
        (1, 1, 2, 2)
        (1, 1, 2, 3)
        (1, 2, 1, 1)
        (1, 2, 1, 2)
        (1, 2, 1, 3)
        (1, 2, 2, 1)
        (1, 2, 2, 2)
        (1, 2, 2, 3)
        (1, 2, 3, 1)
        (1, 2, 3, 2)
        (1, 2, 3, 3)
        (1, 2, 3, 4)

    Returns restricted growth functions in lex order.

    Returns generator of tuples.
    """
    assert _math.is_positive_integer(length), repr(length)
    last_rgf = list(range(1, length + 1))
    rgf = length * [1]
    yield tuple(rgf)
    while not rgf == last_rgf:
        for i, x in enumerate(reversed(rgf)):
            stop = -(i + 1)
            if x < max(rgf[:stop]) + 1:
                first_part = rgf[:stop]
                increased_part = [rgf[stop] + 1]
                trailing_ones = i * [1]
                rgf = first_part + increased_part + trailing_ones
                yield tuple(rgf)
                break


def outer_product(argument):
    """
    Yields outer product of sequences in ``argument``.

    ..  container:: example

        >>> sequences = [list([1, 2, 3]), list(["a", "b"])]
        >>> for sequence in abjad.enumerate.outer_product(sequences):
        ...     sequence
        ...
        [1, 'a']
        [1, 'b']
        [2, 'a']
        [2, 'b']
        [3, 'a']
        [3, 'b']

    ..  container:: example

        >>> sequences = [[1, 2, 3], ["a", "b"], ["X", "Y"]]
        >>> for sequence in abjad.enumerate.outer_product(sequences):
        ...     sequence
        ...
        [1, 'a', 'X']
        [1, 'a', 'Y']
        [1, 'b', 'X']
        [1, 'b', 'Y']
        [2, 'a', 'X']
        [2, 'a', 'Y']
        [2, 'b', 'X']
        [2, 'b', 'Y']
        [3, 'a', 'X']
        [3, 'a', 'Y']
        [3, 'b', 'X']
        [3, 'b', 'Y']

    ..  container:: example

        >>> sequences = [[1, 2, 3], [4, 5], [6, 7, 8]]
        >>> for sequence in abjad.enumerate.outer_product(sequences):
        ...     sequence
        ...
        [1, 4, 6]
        [1, 4, 7]
        [1, 4, 8]
        [1, 5, 6]
        [1, 5, 7]
        [1, 5, 8]
        [2, 4, 6]
        [2, 4, 7]
        [2, 4, 8]
        [2, 5, 6]
        [2, 5, 7]
        [2, 5, 8]
        [3, 4, 6]
        [3, 4, 7]
        [3, 4, 8]
        [3, 5, 6]
        [3, 5, 7]
        [3, 5, 8]

    Returns list of lists.
    """

    def _helper(sequence_1, sequence_2):
        result = []
        for item_1 in sequence_1:
            for item_2 in sequence_2:
                result.extend([item_1 + [item_2]])
        return result

    argument = list(argument)
    argument[0] = [[_] for _ in argument[0]]
    result = functools.reduce(_helper, argument)
    return result


def yield_combinations(argument, minimum_length=None, maximum_length=None):
    """
    Yields combinations of sequence items.

    ..  container:: example

        >>> sequence = list([1, 2, 3, 4])
        >>> for combination in abjad.enumerate.yield_combinations(sequence):
        ...     combination
        []
        [1]
        [2]
        [1, 2]
        [3]
        [1, 3]
        [2, 3]
        [1, 2, 3]
        [4]
        [1, 4]
        [2, 4]
        [1, 2, 4]
        [3, 4]
        [1, 3, 4]
        [2, 3, 4]
        [1, 2, 3, 4]

    ..  container:: example

        >>> sequence = list([1, 2, 3, 4])
        >>> for combination in abjad.enumerate.yield_combinations(
        ...     sequence,
        ...     minimum_length=3,
        ...     ):
        ...     combination
        [1, 2, 3]
        [1, 2, 4]
        [1, 3, 4]
        [2, 3, 4]
        [1, 2, 3, 4]

    ..  container:: example

        >>> sequence = list([1, 2, 3, 4])
        >>> for combination in abjad.enumerate.yield_combinations(
        ...     sequence,
        ...     maximum_length=2,
        ...     ):
        ...     combination
        []
        [1]
        [2]
        [1, 2]
        [3]
        [1, 3]
        [2, 3]
        [4]
        [1, 4]
        [2, 4]
        [3, 4]

    ..  container:: example

        >>> sequence = list([1, 2, 3, 4])
        >>> for combination in abjad.enumerate.yield_combinations(
        ...     sequence,
        ...     minimum_length=2,
        ...     maximum_length=2,
        ...     ):
        ...     combination
        [1, 2]
        [1, 3]
        [2, 3]
        [1, 4]
        [2, 4]
        [3, 4]

    ..  container:: example

        >>> sequence = list("text")
        >>> for combination in abjad.enumerate.yield_combinations(sequence):
        ...     ''.join([str(_) for _ in combination])
        ''
        't'
        'e'
        'te'
        'x'
        'tx'
        'ex'
        'tex'
        't'
        'tt'
        'et'
        'tet'
        'xt'
        'txt'
        'ext'
        'text'

    Yields combinations in binary string order.

    Returns sequence generator.
    """
    length = len(argument)
    for i in range(2**length):
        binary_string = _math.integer_to_binary_string(i)
        binary_string = binary_string.zfill(length)
        sublist = []
        for j, digit in enumerate(reversed(binary_string)):
            if digit == "1":
                sublist.append(argument[j])
        yield_sublist = True
        if minimum_length is not None:
            if len(sublist) < minimum_length:
                yield_sublist = False
        if maximum_length is not None:
            if maximum_length < len(sublist):
                yield_sublist = False
        if yield_sublist:
            if isinstance(argument, str):
                yield "".join(sublist)
            else:
                yield type(argument)(sublist)


def yield_pairs(argument):
    """
    Yields pairs sequence items.

    ..  container:: example

        Without duplicate items:

        >>> for pair in abjad.enumerate.yield_pairs([1, 2, 3, 4]):
        ...     pair
        ...
        [1, 2]
        [1, 3]
        [1, 4]
        [2, 3]
        [2, 4]
        [3, 4]

    ..  container:: example

        With duplicate items:

        >>> for pair in abjad.enumerate.yield_pairs([1, 1, 1]):
        ...     pair
        ...
        [1, 1]
        [1, 1]
        [1, 1]

    ..  container:: example

        Length-1 sequence:

        >>> for pair in abjad.enumerate.yield_pairs([1]):
        ...     pair
        ...

    ..  container:: example

        Empty sequence:

        >>> for pair in abjad.enumerate.yield_pairs([]):
        ...     pair
        ...

    Returns generator of length-2 sequences.
    """
    for i, item in enumerate(argument):
        start = i + 1
        for item_ in argument[start:]:
            pair = [item, item_]
            yield type(argument)(pair)


def yield_partitions(argument):
    """
    Yields partitions of sequence.

    ..  container:: example

        >>> for partition in abjad.enumerate.yield_partitions([0, 1, 2]):
        ...     partition
        ...
        [[0, 1, 2]]
        [[0, 1], [2]]
        [[0], [1, 2]]
        [[0], [1], [2]]

    ..  container:: example

        >>> for partition in abjad.enumerate.yield_partitions([0, 1, 2, 3]):
        ...     partition
        ...
        [[0, 1, 2, 3]]
        [[0, 1, 2], [3]]
        [[0, 1], [2, 3]]
        [[0, 1], [2], [3]]
        [[0], [1, 2, 3]]
        [[0], [1, 2], [3]]
        [[0], [1], [2, 3]]
        [[0], [1], [2], [3]]

    Returns generator of nested sequences.
    """
    length = len(argument) - 1
    for i in range(2**length):
        binary_string = _math.integer_to_binary_string(i)
        binary_string = binary_string.zfill(length)
        part = list(argument[0:1])
        partition = [part]
        for n, token in zip(argument[1:], binary_string):
            if int(token) == 0:
                part.append(n)
            else:
                part = [n]
                partition.append(part)
        parts = [type(argument)(_) for _ in partition]
        partition = type(argument)(parts)
        yield partition


def yield_permutations(argument) -> typing.Generator[typing.Any, None, None]:
    """
    Yields permutations of sequence.

    ..  container:: example

        >>> for permutation in abjad.enumerate.yield_permutations([1, 2, 3]):
        ...     permutation
        ...
        [1, 2, 3]
        [1, 3, 2]
        [2, 1, 3]
        [2, 3, 1]
        [3, 1, 2]
        [3, 2, 1]

    Returns sequence generator.
    """
    _argument = list(argument)
    length = len(_argument)
    for permutation in itertools.permutations(tuple(range(length))):
        permutation = type(argument)(permutation)
        yield _sequence.permute(argument, permutation)


def yield_set_partitions(argument):
    """
    Yields set partitions of sequence.

    ..  container:: example

        >>> for set_partition in abjad.enumerate.yield_set_partitions([21, 22, 23, 24]):
        ...     set_partition
        ...
        [[21, 22, 23, 24]]
        [[21, 22, 23], [24]]
        [[21, 22, 24], [23]]
        [[21, 22], [23, 24]]
        [[21, 22], [23], [24]]
        [[21, 23, 24], [22]]
        [[21, 23], [22, 24]]
        [[21, 23], [22], [24]]
        [[21, 24], [22, 23]]
        [[21], [22, 23, 24]]
        [[21], [22, 23], [24]]
        [[21, 24], [22], [23]]
        [[21], [22, 24], [23]]
        [[21], [22], [23, 24]]
        [[21], [22], [23], [24]]

    Returns set partitions in order of restricted growth function.

    Returns generator of list of lists.
    """
    _argument = list(argument)
    length = len(_argument)
    for rgf in _yield_restricted_growth_functions(length):
        partition = _partition_by_rgf(_argument, rgf)
        yield partition


def yield_subsequences(argument, minimum_length=0, maximum_length=None):
    """
    Yields subsequences of ``sequence``.

    ..  container:: example

        >>> for subsequence in abjad.enumerate.yield_subsequences([0, 1, 2]):
        ...     subsequence
        ...
        []
        [0]
        [0, 1]
        [0, 1, 2]
        [1]
        [1, 2]
        [2]

    ..  container:: example

        >>> for subsequence in abjad.enumerate.yield_subsequences(
        ...     [0, 1, 2, 3, 4],
        ...     minimum_length=3,
        ...     ):
        ...     subsequence
        ...
        [0, 1, 2]
        [0, 1, 2, 3]
        [0, 1, 2, 3, 4]
        [1, 2, 3]
        [1, 2, 3, 4]
        [2, 3, 4]

    ..  container:: example

        >>> for subsequence in abjad.enumerate.yield_subsequences(
        ...     [0, 1, 2, 3, 4],
        ...     maximum_length=3,
        ...     ):
        ...     subsequence
        ...
        []
        [0]
        [0, 1]
        [0, 1, 2]
        [1]
        [1, 2]
        [1, 2, 3]
        [2]
        [2, 3]
        [2, 3, 4]
        [3]
        [3, 4]
        [4]

    ..  container:: example

        >>> for subsequence in abjad.enumerate.yield_subsequences(
        ...     [0, 1, 2, 3, 4],
        ...     minimum_length=3,
        ...     maximum_length=3,
        ...     ):
        ...     subsequence
        ...
        [0, 1, 2]
        [1, 2, 3]
        [2, 3, 4]

    Returns sequence generator.
    """
    _argument = argument
    length = len(_argument)
    if maximum_length is None:
        maximum_length = length
    for i in range(length):
        start_j = minimum_length + i
        stop_j = min(maximum_length + i, length) + 1
        for j in range(start_j, stop_j):
            if i < j or i == 0:
                subsequence = _argument[i:j]
                yield subsequence



import pytest

import abjad


def test_Leaf_duration_assign_01():
    """
    Written duration can be assigned a duration.
    """

    note = abjad.Note(1, (1, 4))
    note.written_duration = abjad.Duration(1, 8)
    assert note.written_duration == abjad.Duration(1, 8)


def test_Leaf_duration_assign_02():
    """
    Written duration can be assigned an integer.
    """

    note = abjad.Note(1, (1, 4))
    note.written_duration = 2
    assert note.written_duration == abjad.Duration(2, 1)


def test_Leaf_duration_assign_03():
    """
    Written duration can be assigned an tuple.
    """

    note = abjad.Note(1, (1, 4))
    note.written_duration = (1, 2)
    assert note.written_duration == abjad.Duration(1, 2)


def test_Leaf_duration_compare_01():
    """
    Written durations can be evaluated for equality with durations.
    """

    note = abjad.Note("c'4")
    assert note.written_duration == abjad.Duration(1, 4)


def test_Leaf_duration_compare_02():
    """
    Written durations can be evaluated for equality with integers.
    """

    note = abjad.Note(0, 1)
    assert note.written_duration == 1


def test_Leaf_duration_compare_03():
    """
    Written durations can NOT be evaluated for equality with tuples.
    """

    note = abjad.Note("c'4")
    assert note.written_duration == abjad.Duration(1, 4)
    assert note.written_duration != (1, 4)
    assert note.written_duration != "foo"


def test_Leaf_written_duration_01():
    """
    Leaf durations can go up to 'maxima...': duration < (16, 1).
    """

    note = abjad.Note(1, 2)

    assert abjad.lilypond(note) == "cs'\\breve"
    note.written_duration = abjad.Duration(3)
    assert abjad.lilypond(note) == "cs'\\breve."
    note.written_duration = abjad.Duration(4)
    assert abjad.lilypond(note) == "cs'\\longa"
    note.written_duration = abjad.Duration(6)
    assert abjad.lilypond(note) == "cs'\\longa."
    note.written_duration = abjad.Duration(7)
    assert abjad.lilypond(note) == "cs'\\longa.."
    note.written_duration = abjad.Duration(8)
    assert abjad.lilypond(note) == "cs'\\maxima"
    note.written_duration = abjad.Duration(12)
    assert abjad.lilypond(note) == "cs'\\maxima."
    note.written_duration = abjad.Duration(14)
    assert abjad.lilypond(note) == "cs'\\maxima.."
    note.written_duration = abjad.Duration(15)
    assert abjad.lilypond(note) == "cs'\\maxima..."

    with pytest.raises(abjad.AssignabilityError):
        abjad.Note(1, 16)



import abjad


def test_AfterGraceContainer_01():
    r"""
    REGRESSION. LilyPond \afterGrace must lexically precede LilyPond \pitchedTrill.
    """

    voice = abjad.Voice("c'1")
    container = abjad.IndependentAfterGraceContainer("e'8")
    voice.append(container)
    voice.append("r4")
    start_trill_span = abjad.StartTrillSpan(pitch=abjad.NamedPitch("D4"))
    abjad.attach(start_trill_span, voice[0])
    stop_trill_span = abjad.StopTrillSpan()
    abjad.attach(stop_trill_span, voice[-1])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \afterGrace
            \pitchedTrill
            c'1
            \startTrillSpan d'
            {
                e'8
            }
            r4
            \stopTrillSpan
        }
        """
    )



"""
Abjad's math library.
"""

import collections
import fractions
import itertools
import math
import numbers
import typing


def all_are_equal(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of equal items.

    ..  container:: example

        >>> abjad.math.all_are_equal([99, 99, 99, 99, 99, 99])
        True

        >>> abjad.math.all_are_equal(17)
        False

    ..  container:: example

        Is true when ``argument`` is empty:

        >>> abjad.math.all_are_equal([])
        True

    """
    try:
        first_element = None
        for element in argument:
            if first_element is None:
                first_element = element
            else:
                if not element == first_element:
                    return False
        return True
    except TypeError:
        return False


def all_are_integer_equivalent(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection with
    integer-equivalent items.

    ..  container:: example

        >>> import fractions
        >>> items = [1, '2', 3.0, fractions.Fraction(4, 1)]
        >>> abjad.math.all_are_integer_equivalent(items)
        True

        >>> abjad.math.all_are_integer_equivalent([1, '2', 3.5, 4])
        False

    """
    try:
        return all(is_integer_equivalent(_) for _ in argument)
    except TypeError:
        return False


def all_are_integer_equivalent_numbers(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection with
    integer-equivalent items.

    ..  container:: example

        >>> import fractions
        >>> items = [1, 2, 3.0, fractions.Fraction(4, 1)]
        >>> abjad.math.all_are_integer_equivalent_numbers(items)
        True

        >>> abjad.math.all_are_integer_equivalent_numbers([1, 2, 3.5, 4])
        False

    """
    try:
        return all(is_integer_equivalent_number(_) for _ in argument)
    except TypeError:
        return False


def all_are_nonnegative_integer_equivalent_numbers(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integer-equivalent numbers.

    ..  container:: example

        >>> import fractions
        >>> items = [0, 0.0, fractions.Fraction(0), 2, 2.0, fractions.Fraction(2)]
        >>> abjad.math.all_are_nonnegative_integer_equivalent_numbers(items)
        True

        >>> items = [0, 0.0, fractions.Fraction(0), -2, 2.0, fractions.Fraction(2)]
        >>> abjad.math.all_are_nonnegative_integer_equivalent_numbers(items)
        False

    """
    try:
        return all(is_nonnegative_integer_equivalent_number(_) for _ in argument)
    except TypeError:
        return False


def all_are_nonnegative_integer_powers_of_two(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integer powers of two.

    ..  container:: example

        >>> items = [0, 1, 1, 1, 2, 4, 32, 32]
        >>> abjad.math.all_are_nonnegative_integer_powers_of_two(items)
        True

        >>> abjad.math.all_are_nonnegative_integer_powers_of_two(17)
        False

    ..  container:: example

        Is true when ``argument`` is empty:

        >>> abjad.math.all_are_nonnegative_integer_powers_of_two([])
        True

    """
    try:
        return all(is_nonnegative_integer_power_of_two(_) for _ in argument)
    except TypeError:
        return False


def all_are_nonnegative_integers(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integers.

    ..  container:: example

        >>> abjad.math.all_are_nonnegative_integers([0, 1, 2, 99])
        True

        >>> abjad.math.all_are_nonnegative_integers([0, 1, 2, -99])
        False

    """
    try:
        return all(is_nonnegative_integer(_) for _ in argument)
    except TypeError:
        return False


def all_are_pairs_of_types(argument, first_type, second_type) -> bool:
    """
    Is true when ``argument`` is an iterable collection whose members are all
    of length 2, and where the first member of each pair is an instance of
    ``first_type`` and where the second member of each pair is an instance of
    ``second_type``.

    ..  container:: example

        >>> items = [(1., 'a'), (2.1, 'b'), (3.45, 'c')]
        >>> abjad.math.all_are_pairs_of_types(items, float, str)
        True

        >>> abjad.math.all_are_pairs_of_types('foo', float, str)
        False

    ..  container:: example

        Is true when ``argument`` is empty:

        >>> abjad.math.all_are_pairs_of_types([], float, str)
        True

    """
    try:
        return all(
            (
                len(_) == 2
                and isinstance(_[0], first_type)
                and isinstance(_[1], second_type)
            )
            for _ in argument
        )
    except (KeyError, TypeError):
        return False


def all_are_positive_integers(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of positive integers.

    ..  container:: example

        >>> abjad.math.all_are_positive_integers([1, 2, 3, 99])
        True

        >>> abjad.math.all_are_positive_integers(17)
        False

    """
    try:
        return all(is_positive_integer(_) for _ in argument)
    except TypeError:
        return False


def are_relatively_prime(argument) -> bool:
    """
    Is true when ``argument`` is an iterable collection of relative primes.

    ..  container:: example

        >>> abjad.math.are_relatively_prime([13, 14, 15])
        True

        >>> abjad.math.are_relatively_prime([13, 14, 15, 16])
        False

        >>> abjad.math.are_relatively_prime('text')
        False

    ..  container:: example

        Returns true when ``argument`` is empty:

        >>> abjad.math.are_relatively_prime([])
        True

    """
    if not isinstance(argument, collections.abc.Iterable):
        return False
    if not all(isinstance(_, numbers.Number) for _ in argument):
        return False
    all_factors: typing.Set[int] = set([])
    for number in argument:
        current_factors = factors(number)
        current_factors_ = set(current_factors)
        if all_factors & current_factors_:
            return False
        all_factors.update(current_factors_)
    return True


def arithmetic_mean(argument) -> int | float | fractions.Fraction:
    """
    Gets arithmetic mean of ``argument``.

    ..  container:: example

        >>> abjad.math.arithmetic_mean([1, 2, 2, 20, 30])
        11

        >>> abjad.math.arithmetic_mean([1, 2, 20])
        Fraction(23, 3)

        >>> abjad.math.arithmetic_mean([2, 2, 20.0])
        8.0

    Raises exception when ``argument`` is not iterable.
    """
    if not isinstance(argument, collections.abc.Sequence):
        raise TypeError(argument)
    total = sum(argument)
    length = len(argument)
    if isinstance(total, float):
        return total / length
    result = fractions.Fraction(sum(argument), len(argument))
    int_result = int(result)
    if int_result == result:
        return int_result
    else:
        return result


def binomial_coefficient(n, k) -> int:
    """
    Gets binomial coefficient of ``n`` choose ``k``.

    ..  container:: example

        >>> for k in range(8):
        ...     print(k, '\t', abjad.math.binomial_coefficient(8, k))
        ...
        0  1
        1  8
        2  28
        3  56
        4  70
        5  56
        6  28
        7  8

    """
    return math.factorial(n) // (math.factorial(n - k) * math.factorial(k))


def cumulative_products(argument):
    """
    Gets cumulative products of ``argument``.

    ..  container:: example

        >>> abjad.math.cumulative_products([1, 2, 3, 4, 5, 6, 7, 8])
        [1, 2, 6, 24, 120, 720, 5040, 40320]

        >>> abjad.math.cumulative_products([1, -2, 3, -4, 5, -6, 7, -8])
        [1, -2, -6, 24, 120, -720, -5040, 40320]

    Raises exception when ``argument`` is not iterable.

    Returns new object of ``argument`` type.
    """
    if not isinstance(argument, collections.abc.Iterable):
        raise TypeError(argument)
    if len(argument) == 0:
        return type(argument)([])
    result = [argument[0]]
    for element in argument[1:]:
        result.append(result[-1] * element)
    return type(argument)(result)


def cumulative_sums(argument, start=0):
    """
    Gets cumulative sums of ``argument``.

    ..  container:: example

        >>> abjad.math.cumulative_sums([1, 2, 3, 4, 5, 6, 7, 8], start=0)
        [0, 1, 3, 6, 10, 15, 21, 28, 36]

        >>> abjad.math.cumulative_sums([1, 2, 3, 4, 5, 6, 7, 8], start=None)
        [1, 3, 6, 10, 15, 21, 28, 36]

    Raises exception when ``argument`` is not iterable.

    Returns new object of ``argument`` type.
    """
    if start is None:
        result = []
    else:
        result = [start]
    for item in argument:
        if result:
            new = result[-1] + item
            result.append(new)
        else:
            result.append(item)
    return type(argument)(result)


def difference_series(argument):
    """
    Gets difference series of ``argument``.

    ..  container:: example

        >>> abjad.math.difference_series([1, 1, 2, 3, 5, 5, 6])
        [0, 1, 1, 2, 0, 1]

        >>> abjad.math.difference_series([9, 6, 8, 5, 7, 4, 6])
        [-3, 2, -3, 2, -3, 2]

    Returns new object of ``argument`` type.
    """
    result = []
    for i, n in enumerate(argument[1:]):
        result.append(n - argument[i])
    return type(argument)(result)


def divide_integer_by_ratio(n, ratio) -> list[fractions.Fraction | float]:
    """
    Divides integer ``n`` by tuple ``ratio``.

    ..  container:: example

        >>> abjad.math.divide_integer_by_ratio(1, (1, 1, 3))
        [Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)]

    ..  container:: example

        >>> abjad.math.divide_integer_by_ratio(1.0, (1, 1, 3))
        [0.2, 0.2, 0.6]

    """
    assert isinstance(n, int | float), repr(n)
    assert isinstance(ratio, tuple), repr(ratio)
    denominator = sum(ratio)
    factors = [fractions.Fraction(_, denominator) for _ in ratio]
    result = [n * _ for _ in factors]
    return result


def divisors(n) -> list[int]:
    """
    Gets positive divisors of ``n`` in increasing order.

    ..  container:: example

        >>> abjad.math.divisors(84)
        [1, 2, 3, 4, 6, 7, 12, 14, 21, 28, 42, 84]

        >>> for x in range(10, 20):
        ...     print(x, abjad.math.divisors(x))
        ...
        10 [1, 2, 5, 10]
        11 [1, 11]
        12 [1, 2, 3, 4, 6, 12]
        13 [1, 13]
        14 [1, 2, 7, 14]
        15 [1, 3, 5, 15]
        16 [1, 2, 4, 8, 16]
        17 [1, 17]
        18 [1, 2, 3, 6, 9, 18]
        19 [1, 19]

    ..  container:: example

        Allows nonpositive ``n``:

        >>> abjad.math.divisors(-27)
        [1, 3, 9, 27]

    Raises not implemented error on ``0``.
    """
    if not isinstance(n, int):
        raise TypeError(f"must be integer: {n!r}.")
    if n == 0:
        raise NotImplementedError("all numbers divide zero evenly.")
    n = abs(n)
    divisors = [1]
    for i in range(2, int(math.sqrt(n)) + 1):
        if n % i == 0:
            divisors.append(i)
    codivisors = [n // i for i in reversed(divisors)]
    if divisors[-1] == codivisors[0]:
        divisors.pop()
    divisors.extend(codivisors)
    divisors.sort()
    return divisors


def factors(n) -> list[int]:
    """
    Gets prime factors less than or equal to ``n`` .

    ..  container:: example

        >>> abjad.math.factors(84)
        [2, 2, 3, 7]

        >>> for n in range(10, 20):
        ...   print(n, abjad.math.factors(n))
        ...
        10 [2, 5]
        11 [11]
        12 [2, 2, 3]
        13 [13]
        14 [2, 7]
        15 [3, 5]
        16 [2, 2, 2, 2]
        17 [17]
        18 [2, 3, 3]
        19 [19]

    ``n`` must be a positive integer.

    Returns factors in increasing order.
    """
    if not is_positive_integer(n):
        raise TypeError(f"must be positive integer: {n!r}.")
    factor = 2
    factors = []
    while 1 < n:
        if n % factor == 0:
            factors.append(factor)
            n = n / factor
        else:
            factor = factor + 1
    return factors


def fraction_to_proper_fraction(
    rational,
) -> tuple[int, fractions.Fraction]:
    """
    Changes ``rational`` to proper fraction.

    ..  container:: example

        >>> import fractions
        >>> abjad.math.fraction_to_proper_fraction(fractions.Fraction(116, 8))
        (14, Fraction(1, 2))

    """
    assert isinstance(rational, fractions.Fraction), repr(rational)
    quotient = int(rational)
    residue = rational - quotient
    return quotient, residue


def greatest_common_divisor(*integers) -> int:
    """
    Calculates greatest common divisor of ``integers``.

    ..  container:: example

        >>> abjad.math.greatest_common_divisor(84, -94, -144)
        2

    Allows nonpositive input.

    Raises not implemented error when zero is included in input.
    """
    common_divisors = None
    for positive_integer in integers:
        all_divisors = set(divisors(positive_integer))
        if common_divisors is None:
            common_divisors = all_divisors
        else:
            common_divisors &= all_divisors
            if common_divisors == set([1]):
                return 1
    assert isinstance(common_divisors, set)
    return max(common_divisors)


def greatest_power_of_two_less_equal(n, i=0) -> int:
    r"""
    Gets greatest integer power of two less than or equal to positive ``n``.

    ..  container:: example

        >>> for n in range(10, 20):
        ...     print('\t%s\t%s' % (n, abjad.math.greatest_power_of_two_less_equal(n)))
        ...
        10 8
        11 8
        12 8
        13 8
        14 8
        15 8
        16 16
        17 16
        18 16
        19 16

        Greatest-but-``i`` integer power of ``2`` less than or equal to
        positive ``n``:

        >>> for n in range(10, 20):
        ...     print('\t%s\t%s' % (n, abjad.math.greatest_power_of_two_less_equal(n, i=1)))
        ...
        10 4
        11 4
        12 4
        13 4
        14 4
        15 4
        16 8
        17 8
        18 8
        19 8

    """
    if n <= 0:
        raise ValueError(f"must be positive: {n!r}.")
    return 2 ** (int(math.log(n, 2)) - i)


def integer_equivalent_number_to_integer(number) -> int | float:
    """
    Changes integer-equivalent ``number`` to integer.

    ..  container:: example

        Returns integer-equivalent number as integer:

        >>> abjad.math.integer_equivalent_number_to_integer(17.0)
        17

    ..  container:: example

        Returns noninteger-equivalent number unchanged:

        >>> abjad.math.integer_equivalent_number_to_integer(17.5)
        17.5

    """
    if is_integer_equivalent_number(number):
        return int(number)
    else:
        return number


def integer_to_base_k_tuple(n, k) -> tuple[int, ...]:
    """
    Changes nonnegative integer ``n`` to base-`k` tuple.

    ..  container:: example

        Gets base-10 digits of 1066:

        >>> abjad.math.integer_to_base_k_tuple(1066, 10)
        (1, 0, 6, 6)

    ..  container:: example

        Gets base-2 digits of 1066:

        >>> abjad.math.integer_to_base_k_tuple(1066, 2)
        (1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0)

    ..  container:: example

        Gets base-26 digits of 1066:

        >>> abjad.math.integer_to_base_k_tuple(1066, 26)
        (1, 15, 0)

    """
    assert isinstance(n, int), repr(n)
    assert 0 <= n, repr(n)
    if n == 0:
        return (0,)
    result = []
    current_exponent = math.trunc(math.log(n, k))
    remainder = n
    while 0 <= current_exponent:
        current_power = k**current_exponent
        current_digit = remainder // current_power
        result.append(current_digit)
        remainder -= current_digit * current_power
        current_exponent -= 1
    return tuple(result)


def integer_to_binary_string(n) -> str:
    r"""
    Changes positive integer ``n`` to binary string.

    ..  container:: example

        >>> for n in range(1, 16 + 1):
        ...     string = abjad.math.integer_to_binary_string(n)
        ...     print(f"{n}\t{string}")
        ...
        1  1
        2  10
        3  11
        4  100
        5  101
        6  110
        7  111
        8  1000
        9  1001
        10 1010
        11 1011
        12 1100
        13 1101
        14 1110
        15 1111
        16 10000

    """
    if n == 0:
        return "0"
    result = bin(abs(n)).lstrip("-0b")
    if n < 0:
        result = "-" + result
    return result


def is_assignable_integer(argument) -> bool:
    r"""
    Is true when ``argument`` is equivalent to an integer that can be written
    without recourse to ties.

    ..  container:: example

        >>> for n in range(0, 16 + 1):
        ...     print('%s\t%s' % (n, abjad.math.is_assignable_integer(n)))
        ...
        0  False
        1  True
        2  True
        3  True
        4  True
        5  False
        6  True
        7  True
        8  True
        9  False
        10 False
        11 False
        12 True
        13 False
        14 True
        15 True
        16 True

    """
    if isinstance(argument, int):
        if 0 < argument:
            if "01" not in integer_to_binary_string(argument):
                return True
    return False


def is_integer_equivalent(argument) -> bool:
    """
    Is true when ``argument`` is an integer-equivalent number.

    ..  container:: example

        >>> abjad.math.is_integer_equivalent(12.0)
        True

        >>> abjad.math.is_integer_equivalent('12')
        True

        >>> abjad.math.is_integer_equivalent('foo')
        False

    """
    if isinstance(argument, numbers.Number):
        return is_integer_equivalent_number(argument)
    try:
        int(argument)
        return True
    except (TypeError, ValueError):
        return False


def is_integer_equivalent_n_tuple(argument, n) -> bool:
    """
    Is true when ``argument`` is a tuple of ``n`` integer-equivalent items.

    ..  container:: example

        >>> import fractions
        >>> tuple_ = (2.0, '3', fractions.Fraction(4, 1))
        >>> abjad.math.is_integer_equivalent_n_tuple(tuple_, 3)
        True

        >>> tuple_ = (2.5, '3', fractions.Fraction(4, 1))
        >>> abjad.math.is_integer_equivalent_n_tuple(tuple_, 3)
        False

    """
    return (
        isinstance(argument, tuple)
        and len(argument) == n
        and all(is_integer_equivalent(_) for _ in argument)
    )


def is_integer_equivalent_number(argument) -> bool:
    """
    Is true when ``argument`` is a number and ``argument`` is equivalent to an
    integer.

    ..  container:: example

        >>> abjad.math.is_integer_equivalent_number(12.0)
        True

        >>> abjad.math.is_integer_equivalent_number(abjad.Duration(1, 2))
        False

    """
    if int(argument) == argument:
        return True
    return False


def is_nonnegative_integer(argument) -> bool:
    """
    Is true when ``argument`` equals a nonnegative integer.

    ..  container:: example

        >>> abjad.math.is_nonnegative_integer(99)
        True

        >>> abjad.math.is_nonnegative_integer(0)
        True

        >>> abjad.math.is_nonnegative_integer(-1)
        False

    """
    if argument == int(argument):
        if 0 <= argument:
            return True
    return False


def is_nonnegative_integer_equivalent_number(argument) -> bool:
    """
    Is true when ``argument`` is a nonnegative integer-equivalent number.

    ..  container:: example

        >>> duration = abjad.Duration(4, 2)
        >>> abjad.math.is_nonnegative_integer_equivalent_number(duration)
        True

    """
    return is_integer_equivalent_number(argument) and 0 <= argument


def is_nonnegative_integer_power_of_two(argument) -> bool:
    """
    Is true when ``argument`` is a nonnegative integer power of 2.

    ..  container:: example

        >>> for n in range(10):
        ...     print(n, abjad.math.is_nonnegative_integer_power_of_two(n))
        ...
        0 True
        1 True
        2 True
        3 False
        4 True
        5 False
        6 False
        7 False
        8 True
        9 False

    """
    if isinstance(argument, int):
        return not bool(argument & (argument - 1))
    elif isinstance(argument, fractions.Fraction):
        return is_nonnegative_integer_power_of_two(
            argument.numerator * argument.denominator
        )
    else:
        return False


def is_positive_integer(argument) -> bool:
    """
    Is true when ``argument`` equals a positive integer.

    ..  container:: example

        >>> abjad.math.is_positive_integer(99)
        True

        >>> abjad.math.is_positive_integer(0)
        False

        >>> abjad.math.is_positive_integer(-1)
        False

    """
    if argument == int(argument):
        if 0 < argument:
            return True
    return False


def is_positive_integer_equivalent_number(argument) -> bool:
    """
    Is true when ``argument`` is a positive integer-equivalent number.

    ..  container:: example

        >>> abjad.math.is_positive_integer_equivalent_number(
        ...     abjad.Duration(4, 2)
        ...     )
        True

    """
    try:
        return 0 < argument and is_integer_equivalent_number(argument)
    except TypeError:  # Python 3 comparisons with non-numbers
        return False


def is_positive_integer_power_of_two(argument) -> bool:
    r"""
    Is true when ``argument`` is a positive integer power of 2.

    ..  container:: example

        >>> for n in range(10):
        ...     print(n, abjad.math.is_positive_integer_power_of_two(n))
        ...
        0 False
        1 True
        2 True
        3 False
        4 True
        5 False
        6 False
        7 False
        8 True
        9 False

    """
    return 0 < argument and is_nonnegative_integer_power_of_two(argument)


def least_common_multiple(*integers) -> int:
    """
    Gets least common multiple of positive ``integers``.

    ..  container:: example

        >>> abjad.math.least_common_multiple(2, 4, 5, 10, 20)
        20

        >>> abjad.math.least_common_multiple(4, 4)
        4

        >>> abjad.math.least_common_multiple(4, 5)
        20

        >>> abjad.math.least_common_multiple(4, 6)
        12

        >>> abjad.math.least_common_multiple(4, 7)
        28

        >>> abjad.math.least_common_multiple(4, 8)
        8

        >>> abjad.math.least_common_multiple(4, 9)
        36

        >>> abjad.math.least_common_multiple(4, 10)
        20

        >>> abjad.math.least_common_multiple(4, 11)
        44

    """
    if len(integers) == 1:
        if not isinstance(integers[0], int):
            raise TypeError(f"must be integer: {integers[0]!r}.")
        if not 0 < integers[0]:
            raise ValueError(f"must be positive: {integers[0]!r}.")
        return integers[0]
    current_lcm = _least_common_multiple_helper(*integers[:2])
    for remaining_positive_integer in integers[2:]:
        current_lcm = _least_common_multiple_helper(
            current_lcm, remaining_positive_integer
        )
    return current_lcm


def _least_common_multiple_helper(m, n):
    assert isinstance(m, int), repr(m)
    assert isinstance(n, int), repr(n)
    factors_m = factors(m)
    factors_n = factors(n)
    for x in factors_m:
        try:
            factors_n.remove(x)
        except ValueError:
            pass
    result = 1
    for x in factors_m + factors_n:
        result *= x
    return result


def partition_integer_by_ratio(n, ratio) -> list[int]:
    """
    Partitions positive integer-equivalent ``n`` by ``ratio``.

    Returns result with weight equal to absolute value of ``n``.

    ..  container:: example

        >>> abjad.math.partition_integer_by_ratio(10, (1, 2))
        [3, 7]

    ..  container:: example

        Partitions positive integer-equivalent ``n`` by ``ratio`` with negative
        parts:

        >>> abjad.math.partition_integer_by_ratio(10, (1, -2))
        [3, -7]

    ..  container:: example

        Partitions negative integer-equivalent ``n`` by ``ratio``:

        >>> abjad.math.partition_integer_by_ratio(-10, (1, 2))
        [-3, -7]

    ..  container:: example

        Partitions negative integer-equivalent ``n`` by ``ratio`` with negative
        parts:

        >>> abjad.math.partition_integer_by_ratio(-10, (1, -2))
        [-3, 7]

    ..  container:: example

        More examples:

        >>> abjad.math.partition_integer_by_ratio(10, (1,))
        [10]

        >>> abjad.math.partition_integer_by_ratio(10, (1, 1))
        [5, 5]

        >>> abjad.math.partition_integer_by_ratio(10, (1, -1, -1))
        [3, -4, -3]

        >>> abjad.math.partition_integer_by_ratio(-10, (1, 1, 1, 1))
        [-3, -2, -3, -2]

        >>> abjad.math.partition_integer_by_ratio(-10, (1, 1, 1, 1, 1))
        [-2, -2, -2, -2, -2]

    """
    if not is_integer_equivalent_number(n):
        raise TypeError(f"is not integer-equivalent number: {n!r}.")
    if not all(is_integer_equivalent_number(_) for _ in ratio):
        raise ValueError(f"must be integer tuple ratio, not {ratio!r}.")
    result = [0]
    parts = [float(abs(n)) * abs(_) / weight(ratio) for _ in ratio]
    cumulative_parts = cumulative_sums(parts, start=None)
    for part in cumulative_parts:
        rounded_part = int(round(part)) - sum(result)
        if part - round(part) == 0.5:
            rounded_part += 1
        result.append(rounded_part)
    result = result[1:]
    if sign(n) == -1:
        result = [-_ for _ in result]
    ratio_signs = [sign(_) for _ in ratio]
    result = [pair[0] * pair[1] for pair in zip(ratio_signs, result)]
    return result


def partition_integer_into_canonic_parts(
    n, decrease_parts_monotonically=True
) -> tuple[int, ...]:
    """
    Partitions integer ``n`` into canonic parts.

    Returns all parts positive on positive ``n``:

    ..  container:: example

        >>> for n in range(1, 11):
        ...     print(n, abjad.math.partition_integer_into_canonic_parts(n))
        ...
        1 (1,)
        2 (2,)
        3 (3,)
        4 (4,)
        5 (4, 1)
        6 (6,)
        7 (7,)
        8 (8,)
        9 (8, 1)
        10 (8, 2)

    ..  container:: example

        Returns all parts negative on negative ``n``:

        >>> for n in reversed(range(-20, -10)):
        ...     print(n, abjad.math.partition_integer_into_canonic_parts(n))
        ...
        -11 (-8, -3)
        -12 (-12,)
        -13 (-12, -1)
        -14 (-14,)
        -15 (-15,)
        -16 (-16,)
        -17 (-16, -1)
        -18 (-16, -2)
        -19 (-16, -3)
        -20 (-16, -4)

    ..  container:: example

        Returns parts that increase monotonically:

        >>> for n in range(11, 21):
        ...     print(n, abjad.math.partition_integer_into_canonic_parts(n,
        ...         decrease_parts_monotonically=False))
        ...
        11 (3, 8)
        12 (12,)
        13 (1, 12)
        14 (14,)
        15 (15,)
        16 (16,)
        17 (1, 16)
        18 (2, 16)
        19 (3, 16)
        20 (4, 16)

    Returns tuple with parts that decrease monotonically.
    """
    assert isinstance(n, int), repr(n)
    assert isinstance(decrease_parts_monotonically, bool)
    if n == 0:
        return (0,)
    result = []
    previous_empty = True
    binary_n = integer_to_binary_string(abs(n))
    binary_length = len(binary_n)
    for i, character in enumerate(binary_n):
        if character == "1":
            place_value = 2 ** (binary_length - i - 1)
            if previous_empty:
                result.append(place_value)
            else:
                result[-1] += place_value
            previous_empty = False
        else:
            previous_empty = True
    sign_n = sign(n)
    if sign(n) == -1:
        result = [sign_n * _ for _ in result]
    if decrease_parts_monotonically:
        return tuple(result)
    else:
        return tuple(reversed(result))


def sign(n) -> int:
    """
    Gets sign of ``n``.

    ..  container:: example

        Returns -1 on negative ``n``:

        >>> abjad.math.sign(-96.2)
        -1

        Returns 0 when ``n`` is 0:

        >>> abjad.math.sign(0)
        0

        Returns 1 on positive ``n``:

        >>> abjad.math.sign(abjad.Duration(9, 8))
        1

    """
    return (0 < n) - (n < 0)


def weight(argument) -> int:
    """
    Gets weight of ``argument``.

    ..  container:: example

        >>> abjad.math.weight([-1, -2, 3, 4, 5])
        15

    ..  container:: example

        >>> abjad.math.weight([])
        0

    Defined equal to sum of the absolute value of items in ``argument``.
    """
    return sum([abs(_) for _ in argument])


def yield_all_compositions_of_integer(n: int) -> typing.Iterator[tuple[int, ...]]:
    """
    Yields all compositions of positive integer ``n``.

    ..  container:: example

        >>> for tuple_ in abjad.math.yield_all_compositions_of_integer(5):
        ...     tuple_
        ...
        (5,)
        (4, 1)
        (3, 2)
        (3, 1, 1)
        (2, 3)
        (2, 2, 1)
        (2, 1, 2)
        (2, 1, 1, 1)
        (1, 4)
        (1, 3, 1)
        (1, 2, 2)
        (1, 2, 1, 1)
        (1, 1, 3)
        (1, 1, 2, 1)
        (1, 1, 1, 2)
        (1, 1, 1, 1, 1)

    Lists parts in descending lex order.

    Parts sum to ``n``.

    Finds small values of ``n`` easily.

    Takes around 4 seconds for ``n`` equal to 17.
    """
    compositions = []
    integer = 0
    string_length = n
    while integer < 2 ** (n - 1):
        binary_string = integer_to_binary_string(integer)
        binary_string = binary_string.zfill(string_length)
        digits = [int(_) for _ in list(binary_string)]
        partition = []
        generator = itertools.groupby(digits, lambda _: _)
        for value, group in generator:
            partition.append(list(group))
        sublengths = [len(part) for part in partition]
        composition = tuple(sublengths)
        compositions.append(composition)
        integer += 1
    for composition in reversed(sorted(compositions)):
        yield composition


class Infinity:
    """
    Infinity.

    ..  container:: example

        All numbers compare less than infinity:

        >>> 9999999 < abjad.Infinity()
        True

        >>> 2**38 < abjad.Infinity()
        True

    ..  container:: example

        Infinity compares equal to itself:

        >>> abjad.Infinity() == abjad.Infinity()
        True

    ..  container:: example

        Negative infinity compares less than infinity:

        >>> abjad.NegativeInfinity() < abjad.Infinity()
        True

    Initializes as a system singleton at start-up.

    Available as a built-in after Abjad starts.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_value",)

    ### INTIALIZER ###

    def __init__(self):
        self._value = float("infinity")

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Compares type.
        """
        return isinstance(argument, type(self))

    def __float__(self):
        """
        Convert infinity to float.

        Returns float.
        """
        return self._value

    def __ge__(self, argument):
        """
        Is true for all values of ``argument``.

        Returns true.
        """
        return self._value >= argument

    def __gt__(self, argument):
        """
        Is true for all noninfinite values of ``argument``.

        Returns true or false.
        """
        return self._value > argument

    def __hash__(self):
        """
        Hashes infinity.
        """
        return hash(self.__class__.__name__ + str(self))

    def __le__(self, argument):
        """
        Is true when ``argument`` is infinite.

        Returns true or false.
        """
        return self._value <= argument

    def __lt__(self, argument):
        """
        Is true for no values of ``argument``.

        Returns true or false.
        """
        return self._value < argument

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        return f"{type(self).__name__}()"

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from infinity.

        Returns infinity or 0 if ``argument`` is also infinity.
        """
        if argument is self:
            return 0
        return self


class NegativeInfinity(Infinity):
    """
    Negative infinity.

    ..  container:: example

        All numbers compare greater than negative infinity:

        >>> abjad.NegativeInfinity() < -9999999
        True

    ..  container:: example

        Negative infinity compares equal to itself:

        >>> abjad.NegativeInfinity() == abjad.NegativeInfinity()
        True

    ..  container:: example

        Negative infinity compares less than infinity:

        >>> abjad.NegativeInfinity() < abjad.Infinity()
        True

    Initializes as a system singleton at start-up.

    Available as a built-in after Abjad start.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        self._value = float("-infinity")



import collections
import dataclasses
import typing


@dataclasses.dataclass(slots=True, unsafe_hash=True)
class CyclicTuple:
    """
    Cyclic tuple.

    ..  container:: example

        Initializes from string:

        >>> tuple_ = abjad.CyclicTuple('abcd')

        >>> tuple_
        CyclicTuple(items=('a', 'b', 'c', 'd'))

        >>> for x in range(8):
        ...     print(x, tuple_[x])
        ...
        0 a
        1 b
        2 c
        3 d
        4 a
        5 b
        6 c
        7 d

    Cyclic tuples overload the item-getting method of built-in tuples.

    Cyclic tuples return a value for any integer index.

    Cyclic tuples otherwise behave exactly like built-in tuples.
    """

    items: typing.Sequence = ()

    def __post_init__(self):
        self.items = tuple(self.items)

    def __contains__(self, item) -> bool:
        """
        Is true when cyclic tuple contains ``item``.

        ..  container:: example

            >>> tuple_ = abjad.CyclicTuple('abcd')
            >>> 'a' in tuple_
            True

        """
        return self.items.__contains__(item)

    def __eq__(self, argument) -> bool:
        """
        Compares ``items``.
        """
        if isinstance(argument, tuple):
            return self.items == argument
        elif isinstance(argument, type(self)):
            return self.items == argument.items
        return False

    def __getitem__(self, argument) -> typing.Any:
        """
        Gets item or slice identified by ``argument``.

        ..  container:: example

            Gets slice open at right:

            >>> items = [0, 1, 2, 3, 4, 5]
            >>> tuple_ = abjad.CyclicTuple(items=items)
            >>> tuple_[2:]
            (2, 3, 4, 5)

        ..  container:: example

            Gets slice closed at right:

            >>> items = [0, 1, 2, 3, 4, 5]
            >>> tuple_ = abjad.CyclicTuple(items=items)
            >>> tuple_[:15]
            (0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2)

        Raises index error when ``argument`` can not be found in cyclic tuple.
        """
        if isinstance(argument, slice):
            if (argument.stop is not None and argument.stop < 0) or (
                argument.start is not None and argument.start < 0
            ):
                return self.items.__getitem__(argument)
            else:
                return self._get_slice(argument.start, argument.stop)
        if not self:
            raise IndexError(f"cyclic tuple is empty: {self!r}.")
        argument = argument % len(self)
        return self.items.__getitem__(argument)

    def __iter__(self) -> typing.Iterator:
        """
        Iterates cyclic tuple.

        Iterates items only once.

        Does not iterate infinitely.
        """
        return self.items.__iter__()

    def __len__(self) -> int:
        """
        Gets length of cyclic tuple.
        """
        assert isinstance(self.items, tuple | CyclicTuple), repr(self.items)
        return self.items.__len__()

    def _get_slice(self, start_index, stop_index):
        if stop_index is not None and 1000000 < stop_index:
            stop_index = len(self)
        result = []
        if start_index is None:
            start_index = 0
        if stop_index is None:
            indices = range(start_index, len(self))
        else:
            indices = range(start_index, stop_index)
        result = [self[n] for n in indices]
        return tuple(result)


collections.abc.Sequence.register(CyclicTuple)



import bisect
import collections
import enum
import typing

from . import _updatelib
from . import duration as _duration
from . import exceptions as _exceptions
from . import indicators as _indicators
from . import instruments as _instruments
from . import parentage as _parentage
from . import pitch as _pitch
from . import score as _score
from . import timespan as _timespan
from . import typings as _typings


def _are_logical_voice(components, prototype=None):
    prototype = prototype or (_score.Component,)
    if not isinstance(prototype, tuple):
        prototype = (prototype,)
    assert isinstance(prototype, tuple)
    if len(components) == 0:
        return True
    if all(isinstance(_, prototype) and _._parent is None for _ in components):
        return True
    first = components[0]
    if not isinstance(first, prototype):
        return False
    same_logical_voice = True
    parentage = _parentage.Parentage(first)
    first_logical_voice = parentage.logical_voice()
    for component in components[1:]:
        parentage = _parentage.Parentage(component)
        if parentage.logical_voice() != first_logical_voice:
            same_logical_voice = False
        if not parentage.orphan and not same_logical_voice:
            return False
    return True


def _get_annotation(component, annotation, default=None, unwrap: bool = True):
    assert isinstance(annotation, str | enum.Enum), repr(annotation)
    for wrapper in _get_annotation_wrappers(component):
        if wrapper.annotation == annotation:
            if unwrap is True:
                return wrapper.get_item()
            else:
                return wrapper
    return default


def _get_annotation_wrappers(argument):
    result = []
    for wrapper in getattr(argument, "_wrappers", []):
        if wrapper.annotation:
            result.append(wrapper)
    return result


def _get_duration(argument, *, in_seconds: bool = False, preprolated: bool = False):
    if preprolated is True:
        if hasattr(argument, "_get_preprolated_duration"):
            duration = argument._get_preprolated_duration()
        else:
            duration = sum(_._get_preprolated_duration() for _ in argument)
        return duration
    if isinstance(argument, _score.Component):
        if in_seconds is True:
            return _get_duration_in_seconds(argument)
        else:
            return argument._get_duration()
    durations = [_get_duration(_, in_seconds=in_seconds) for _ in argument]
    return _duration.Duration(sum(durations))


def _get_duration_in_seconds(component):
    if isinstance(component, _score.Container):
        if component.simultaneous:
            return max(
                [_duration.Duration(0)]
                + [_get_duration_in_seconds(_) for _ in component]
            )
        else:
            duration = _duration.Duration(0)
            for component_ in component:
                duration += _get_duration_in_seconds(component_)
            return duration
    else:
        mark = _get_effective(component, _indicators.MetronomeMark)
        if mark is not None and not mark.is_imprecise:
            result = (
                component._get_duration()
                / mark.reference_duration
                / mark.units_per_minute
                * 60
            )
            return _duration.Duration(result)
        raise _exceptions.MissingMetronomeMarkError


def _get_effective(
    component, prototype, *, attributes=None, command=None, n=0, unwrap=True
):
    _updatelib._update_now(component, indicators=True)
    candidate_wrappers = {}
    parentage = component._get_parentage()
    enclosing_voice_name = None
    for component_ in parentage:
        if isinstance(component_, _score.Voice):
            if (
                enclosing_voice_name is not None
                and component_.name != enclosing_voice_name
            ):
                continue
            else:
                enclosing_voice_name = component_.name or id(component_)
        local_wrappers = []
        for wrapper in component_._wrappers:
            if wrapper.annotation:
                continue
            if isinstance(wrapper.unbundle_indicator(), prototype):
                append_wrapper = True
                if (
                    command is not None
                    and wrapper.unbundle_indicator().command != command
                ):
                    continue
                if attributes is not None:
                    for name, value in attributes.items():
                        if getattr(wrapper.unbundle_indicator(), name, None) != value:
                            append_wrapper = False
                if not append_wrapper:
                    continue
                local_wrappers.append(wrapper)
        # active indicator takes precendence over inactive indicator
        if any(_.deactivate is True for _ in local_wrappers) and not all(
            _.deactivate is True for _ in local_wrappers
        ):
            local_wrappers = [_ for _ in local_wrappers if _.deactivate is not True]
        for wrapper in local_wrappers:
            offset = wrapper.start_offset
            candidate_wrappers.setdefault(offset, []).append(wrapper)
        if not isinstance(component_, _score.Context):
            continue
        for wrapper in component_._dependent_wrappers:
            if wrapper.annotation:
                continue
            if isinstance(wrapper.unbundle_indicator(), prototype):
                append_wrapper = True
                if (
                    command is not None
                    and wrapper.unbundle_indicator().command != command
                ):
                    continue
                if attributes is not None:
                    for name, value in attributes.items():
                        if getattr(wrapper.unbundle_indicator(), name, None) != value:
                            append_wrapper = False
                if not append_wrapper:
                    continue
                offset = wrapper.start_offset
                candidate_wrappers.setdefault(offset, []).append(wrapper)
    if not candidate_wrappers:
        return
    all_offsets = sorted(candidate_wrappers)
    start_offset = component._get_timespan().start_offset
    index = bisect.bisect(all_offsets, start_offset) - 1 + int(n)
    if index < 0:
        return
    elif len(candidate_wrappers) <= index:
        return
    wrapper = candidate_wrappers[all_offsets[index]][0]
    if unwrap:
        return wrapper.unbundle_indicator()
    return wrapper


def _get_grace_container(component):
    # _score.IndependentAfterGraceContainer is excluded here;
    # exclusion allows iteration to work correctly
    prototype = (
        _score.AfterGraceContainer,
        _score.BeforeGraceContainer,
    )
    for component_ in component._get_parentage():
        if isinstance(component_, prototype):
            return True
        if component_.__class__.__name__ == "OnBeatGraceContainer":
            return True
    return False


def _get_indicator(
    component: _score.Component,
    prototype: _typings.Prototype | None = None,
    *,
    default: typing.Any | None = None,
    unwrap: bool = True,
) -> typing.Any:
    if not isinstance(component, _score.Component):
        raise Exception("can only get indicator on component.")
    indicators = component._get_indicators(prototype=prototype, unwrap=unwrap)
    if not indicators:
        return default
    elif len(indicators) == 1:
        return list(indicators)[0]
    else:
        name = getattr(prototype, "__name__", "")
        strings = "\n".join(["    " + str(_) for _ in indicators])
        string = f"{len(indicators)} {name} indicators attached to {component}:"
        string += f"\n{strings}"
        raise Exception(string)


def _get_leaf_from_leaf(leaf, n):
    assert n in (-1, 0, 1), repr(n)
    if n == 0:
        return leaf
    sibling = leaf._sibling(n)
    if sibling is None:
        return None
    if n == 1:
        components = sibling._get_descendants_starting_with()
    else:
        assert n == -1
        if (
            isinstance(sibling, _score.Container)
            and len(sibling) == 2
            and any(hasattr(_, "_grace_leaf_duration") for _ in sibling)
        ):
            if sibling[0].__class__.__name__ == "OnBeatGraceContainer":
                main_voice = sibling[1]
            else:
                main_voice = sibling[0]
            return main_voice[-1]
        components = sibling._get_descendants_stopping_with()
    for component in components:
        if not isinstance(component, _score.Leaf):
            continue
        if _are_logical_voice([leaf, component]):
            return component


def _get_persistent_wrappers(*, dependent_wrappers=None, omit_with_indicator=None):
    wrappers = {}
    for wrapper in dependent_wrappers:
        if wrapper.annotation:
            continue
        if not getattr(wrapper.unbundle_indicator(), "persistent", False):
            continue
        assert isinstance(wrapper.unbundle_indicator().persistent, bool)
        should_omit = False
        if omit_with_indicator is not None:
            for component in wrapper.component._get_parentage():
                if component._has_indicator(omit_with_indicator):
                    should_omit = True
                    continue
        if should_omit:
            continue
        if hasattr(wrapper.unbundle_indicator(), "parameter"):
            key = wrapper.unbundle_indicator().parameter
        elif isinstance(wrapper.unbundle_indicator(), _instruments.Instrument):
            key = "Instrument"
        else:
            key = str(type(wrapper.unbundle_indicator()))
        if key not in wrappers:
            wrappers[key] = wrapper
        elif (
            wrappers[key].site_adjusted_start_offset
            < wrapper.site_adjusted_start_offset
        ):
            wrappers[key] = wrapper
        elif (
            wrappers[key].site_adjusted_start_offset
            == wrapper.site_adjusted_start_offset
        ):
            if isinstance(
                wrappers[key].unbundle_indicator(), _indicators.StartHairpin
            ) and isinstance(wrapper.unbundle_indicator(), _indicators.Dynamic):
                pass
            elif (
                getattr(wrapper.unbundle_indicator(), "spanner_start", False) is True
                or getattr(wrapper.unbundle_indicator(), "spanner_stop", False) is True
                or getattr(wrapper.unbundle_indicator(), "trend", False) is True
            ):
                wrappers[key] = wrapper
    return wrappers


def _get_sounding_pitch(note):
    if "sounding pitch" in note._get_indicators(str):
        return note.written_pitch
    else:
        instrument = _get_effective(note, _instruments.Instrument)
        if instrument:
            sounding_pitch = instrument.middle_c_sounding_pitch
        else:
            sounding_pitch = _pitch.NamedPitch("C4")
        interval = _pitch.NamedPitch("C4") - sounding_pitch
        sounding_pitch = interval.transpose(note.written_pitch)
        return sounding_pitch


def _get_sounding_pitches(chord):
    if "sounding pitch" in chord._get_indicators(str):
        return chord.written_pitches
    else:
        instrument = _get_effective(chord, _instruments.Instrument)
        if instrument:
            sounding_pitch = instrument.middle_c_sounding_pitch
        else:
            sounding_pitch = _pitch.NamedPitch("C4")
        interval = _pitch.NamedPitch("C4") - sounding_pitch
        sounding_pitches = [
            interval.transpose(pitch) for pitch in chord.written_pitches
        ]
        return tuple(sounding_pitches)


def _get_timespan(argument, in_seconds: bool = False):
    if isinstance(argument, _score.Component):
        return argument._get_timespan(in_seconds=in_seconds)
    assert isinstance(argument, collections.abc.Iterable), repr(argument)
    remaining_items = []
    for i, item in enumerate(argument):
        if i == 0:
            first_item = item
        else:
            remaining_items.append(item)
    timespan = _get_timespan(first_item, in_seconds=in_seconds)
    start_offset = timespan.start_offset
    stop_offset = timespan.stop_offset
    for item in remaining_items:
        timespan = _get_timespan(item, in_seconds=in_seconds)
        if timespan.start_offset < start_offset:
            start_offset = timespan.start_offset
        if stop_offset < timespan.stop_offset:
            stop_offset = timespan.stop_offset
    return _timespan.Timespan(start_offset, stop_offset)
